!+
! Module: rgsframesutils.90
!
! Authors:  Peter Videler - SOC ESTEC <peter@videler.nl>
!          
! Description: Subroutines for rgsframes
! 
! Comments: extracted on 02feb2000 from rgsframes_mod.f90
!           Modifications and cleanup by C Erd, 21-Mar-2001
!
! $Id: rgsframesutils.f90,v 1.17 2001/07/24 13:52:04 cerd Exp $
!- 

module rgsframesutils

  use saslib
  use gti_mod

  ! ......................................................
  implicit none 
  private 
  public &
       pixelsPerFrame, &
       findDuplicates, &
       selectedAddnCopyColumn
  ! ......................................................

  interface findDuplicates
     module procedure findDuplicates1, findDuplicates2, findDuplicates3, &
          findDuplicates4, findDuplicates5, findDuplicates6
  end interface

contains

  ! --------------------------------------------------------------------------------
  function pixelsPerFrame(pixelFrame, frame) result(npixels)
    ! --------------------------------------------------------------------------------
    integer, dimension(:), intent(in) :: pixelFrame   ! frame col in PIXELS table
    integer, dimension(:), intent(in) :: frame        ! frame col in FRAMES table
    integer, dimension(size(frame)) :: npixels
    integer :: iframe, i, start_pixelsrow

    npixels = 0
                                !assume that PIXELS table is sorted w.r.t. FRAMEnumber
    start_pixelsrow = lbound(pixelFrame,1)
    Frames: do iframe=1, ubound(frame,1)

       Pixels: do i=start_pixelsrow, ubound(pixelFrame,1)

          if ( pixelFrame(i) > frame(iframe) ) then
             exit Pixels
          else if ( pixelFrame(i) == frame(iframe) ) then
             npixels(iframe) = npixels(iframe) +1
          end if

       end do Pixels
       if ( i < ubound(pixelFrame,1) ) start_pixelsrow = i
          
!       npixels(iframe) = count(pixelFrame == frame(i))
    end do Frames

  end function pixelsPerFrame


  ! --------------------------------------------------------------------------------
  function findDuplicates1(aa)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications of array elements; requires that first array is sorted!

    integer, dimension(:), intent(in) :: aa
    logical, dimension(size(aa)) :: duplicate

    integer, dimension(:), allocatable :: idx
    integer :: i, a1, a2

    duplicate = .false.

    a1 = aa(lbound(aa,1))       !minval(aa)
    a2 = aa(ubound(aa,1))       !maxval(aa)
    allocate( idx( a1:a2 ))
    idx = a1 - 10

    do i=lbound(aa,1),ubound(aa,1)
       if ( idx( aa(i) ) == aa(i) ) then
          duplicate(i) = .true.

          if ( verbosity(appMsg) >= Noisy ) then
             write(errstr,*) 'Duplicate entry found at row', i
             call message(appMsg,Noisy)
          end if
       else
          idx( aa(i) ) = aa(i)
       end if
    end do

    deallocate( idx )
  end function findDuplicates1


  ! --------------------------------------------------------------------------------
  function findDuplicates2(aa,bb)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications of array elements; requires that first array is sorted!

    integer, dimension(:), intent(in) :: aa,bb
    logical, dimension(size(aa)) :: duplicate

    integer, dimension(:), allocatable :: idx
    integer :: i, b1, b2

    duplicate = .false.

    b1 = minval(bb)
    b2 = maxval(bb)
    allocate( idx( b1:b2 ))
    idx = aa(lbound(aa,1)) - 10

    do i=lbound(aa,1),ubound(aa,1)
       if ( idx( bb(i) ) == aa(i) ) then
          duplicate(i) = .true.

          if ( verbosity(appMsg) >= Noisy ) then
             write(errstr,*) 'Duplicate entry found at row', i
             call message(appMsg,Noisy)
          end if
       else
          idx( bb(i) ) = aa(i)
       end if
    end do

    deallocate( idx )
  end function findDuplicates2


  ! --------------------------------------------------------------------------------
  function findDuplicates3(aa,bb,cc)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications of array elements; requires that first array is sorted!

    integer, dimension(:), intent(in) :: aa,bb,cc
    logical, dimension(size(aa)) :: duplicate

    integer, dimension(:,:), allocatable :: idx
    integer :: i, b1, b2, c1, c2

    duplicate = .false.

    b1 = minval(bb)
    b2 = maxval(bb)
    c1 = minval(cc)
    c2 = maxval(cc)
    allocate( idx( b1:b2, c1:c2 ))
    idx = aa(lbound(aa,1)) - 10

    do i=lbound(aa,1),ubound(aa,1)
       if ( idx( bb(i), cc(i) ) == aa(i) ) then
          duplicate(i) = .true.

          if ( verbosity(appMsg) >= Noisy ) then
             write(errstr,*) 'Duplicate entry found at row', i
             call message(appMsg,Noisy)
          end if
       else
          idx( bb(i), cc(i) ) = aa(i)
       end if
    end do

    deallocate( idx )
  end function findDuplicates3


  ! --------------------------------------------------------------------------------
  function findDuplicates4(aa,bb,cc,dd)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications of array elements; requires that first array is sorted!

    integer, dimension(:), intent(in) :: aa,bb,cc,dd
    logical, dimension(size(aa)) :: duplicate

    integer, dimension(:,:,:), allocatable :: idx
    integer :: i, b1, b2, c1, c2, d1, d2

    duplicate = .false.

    b1 = minval(bb)
    b2 = maxval(bb)
    c1 = minval(cc)
    c2 = maxval(cc)
    d1 = minval(dd)
    d2 = maxval(dd)
    allocate( idx( b1:b2, c1:c2, d1:d2 ))
    idx = aa(lbound(aa,1)) - 10

    do i=lbound(aa,1),ubound(aa,1)
       if ( idx( bb(i), cc(i), dd(i) ) == aa(i) ) then
          duplicate(i) = .true.

          if ( verbosity(appMsg) >= Noisy ) then
             write(errstr,*) 'Duplicate entry found at row', i
             call message(appMsg,Noisy)
          end if
       else
          idx( bb(i), cc(i), dd(i) ) = aa(i)
       end if
    end do

    deallocate( idx )
  end function findDuplicates4


  ! --------------------------------------------------------------------------------
  function findDuplicates5(aa,bb,cc,dd,ee)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications of array elements; requires that first array is sorted!

    integer, dimension(:), intent(in) :: aa,bb,cc,dd,ee
    logical, dimension(size(aa)) :: duplicate

    integer, dimension(:,:,:,:), allocatable :: idx
    integer :: i, b1, b2, c1, c2, d1, d2, e1, e2

    duplicate = .false.

    b1 = minval(bb)
    b2 = maxval(bb)
    c1 = minval(cc)
    c2 = maxval(cc)
    d1 = minval(dd)
    d2 = maxval(dd)
    e1 = minval(ee)
    e2 = maxval(ee)
    allocate( idx( b1:b2, c1:c2, d1:d2, e1:e2 ))
    idx = aa(lbound(aa,1)) - 10

    do i=lbound(aa,1),ubound(aa,1)
       if ( idx( bb(i), cc(i), dd(i), ee(i) ) == aa(i) ) then
          duplicate(i) = .true.

          if ( verbosity(appMsg) >= Noisy ) then
             write(errstr,*) 'Duplicate entry found at row', i
             call message(appMsg,Noisy)
          end if
       else
          idx( bb(i), cc(i), dd(i), ee(i) ) = aa(i)
       end if
    end do

    deallocate( idx )
  end function findDuplicates5


  ! --------------------------------------------------------------------------------
  function findDuplicates6(aa,bb,cc,dd,ee,ff)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications of array elements; requires that first array is sorted!

    integer, dimension(:), intent(in) :: aa,bb,cc,dd,ee,ff
    logical, dimension(size(aa)) :: duplicate

    integer, dimension(:,:,:,:,:), allocatable :: idx
    integer :: i, b1, b2, c1, c2, d1, d2, e1, e2, f1, f2

    duplicate = .false.

    b1 = minval(bb)
    b2 = maxval(bb)
    c1 = minval(cc)
    c2 = maxval(cc)
    d1 = minval(dd)
    d2 = maxval(dd)
    e1 = minval(ee)
    e2 = maxval(ee)
    f1 = minval(ff)
    f2 = maxval(ff)
    allocate( idx( b1:b2, c1:c2, d1:d2, e1:e2, f1:f2 ))
    idx = aa(lbound(aa,1)) - 10

    do i=lbound(aa,1),ubound(aa,1)
       if ( idx( bb(i), cc(i), dd(i), ee(i), ff(i) ) == aa(i) ) then
          duplicate(i) = .true.

          if ( verbosity(appMsg) >= Noisy ) then
             write(errstr,*) 'Duplicate entry found at row', i
             call message(appMsg,Noisy)
          end if
       else
          idx( bb(i), cc(i), dd(i), ee(i), ff(i) ) = aa(i)
       end if
    end do

    deallocate( idx )
  end function findDuplicates6


  ! --------------------------------------------------------------------------------
  recursive function findDuplicates_old(aa,bb,cc,dd,ee,ff,gg)  result(duplicate)
    ! --------------------------------------------------------------------------------
    ! find duplications (array does not need to be sorted)
    ! Note: the speed of this search depends on the sequence of arguments given;
    !       pass least frequent occurance first

    integer, dimension(:), intent(in) :: aa
    integer, dimension(:), intent(in), optional :: bb,cc,dd,ee,ff,gg
    logical, dimension(size(aa)) :: duplicate

    integer :: i, ii, n, m, ji
    integer, dimension(size(aa)) :: j, idx
    logical, dimension(size(aa)) :: alreadychecked

    m = size(aa,1)
    alreadychecked = .false.

    duplicate = .false.

    idx = (/ (i,i=1,ubound(aa,1)) /)
    Rows: do i=1, ubound(aa,1)-1

       if ( .not. duplicate(i) .and. .not. alreadychecked(i) ) then !already identified?
          j = -1

          !below replaces for sake of speed:
          !            j = pack( idx(i+1:), aa(i+1:) .eq. aa(i), j )
          ji = 0
          do ii=i+1,min(i+25,m)        !check only a limited range for sake of speed
             if ( aa(ii) .eq. aa(i) ) then
                ji = ji +1
                j(ji) = idx(ii)
             end if
          end do

          n = count( j > -1 )
          if ( n > 0 ) then     !duplications found
             duplicate(j(1:n)) = .true.

             j(2:n+1) = j(1:n)
             j(1) = i           !also need to check the current row
             n = n +1

             if ( present(gg) ) then
                duplicate(j(1:n)) = ( duplicate(j(1:n)) .and. & 
                     findDuplicates_old(bb(j(1:n)),cc(j(1:n)),dd(j(1:n)), &
                                    ee(j(1:n)),ff(j(1:n)),gg(j(1:n))) )
             else if ( present(ff) ) then
                duplicate(j(1:n)) = ( duplicate(j(1:n)) .and. & 
                     findDuplicates_old(bb(j(1:n)),cc(j(1:n)),dd(j(1:n)), &
                                    ee(j(1:n)),ff(j(1:n))) )
             else if ( present(ee) ) then
                duplicate(j(1:n)) = ( duplicate(j(1:n)) .and. & 
                     findDuplicates_old(bb(j(1:n)),cc(j(1:n)),dd(j(1:n)), &
                                    ee(j(1:n))) )
             else if ( present(dd) ) then
                duplicate(j(1:n)) = ( duplicate(j(1:n)) .and. & 
                     findDuplicates_old(bb(j(1:n)),cc(j(1:n)),dd(j(1:n))) )
             else if ( present(cc) ) then
                duplicate(j(1:n)) = ( duplicate(j(1:n)) .and. & 
                     findDuplicates_old(bb(j(1:n)),cc(j(1:n))) )
             else if ( present(bb) ) then
                duplicate(j(1:n)) = ( duplicate(j(1:n)) .and. & 
                     findDuplicates_old(bb(j(1:n))) )
             end if
          end if
       end if
       alreadychecked(j(1:n)) = .true.
    end do Rows

    n = count( duplicate )
    if ( n > 0 ) then
       j = -1
       j = pack( idx, duplicate, j )

       do i=1,n
          write(errstr,*) 'Duplicate entry found at row', j(i)
          call message(appMsg,Noisy)
       end do
    end if
  end function findDuplicates_old


  ! --------------------------------------------------------------------------------
  subroutine selectedAddnCopyColumn (inTable, inColName, idx, &
       outTable, outColName, unit, comment)
    ! --------------------------------------------------------------------------------
    ! copy only IDX-ed elements from column name COLNAME to DATA.

    use dal
    use errorHandling

    type(TableT), intent(in) :: inTable, outTable
    character(len=*), intent(in) :: inColName, outColName
    integer, dimension(:), intent(in) :: idx

    character(len=*), intent(in), optional :: unit, comment

    type(ColumnT) :: incol, outcol
    integer(kind=int32), dimension(:), pointer :: indata_i32, outdata_i32
    integer(kind=int16), dimension(:), pointer :: indata_i16, outdata_i16
    integer(kind=int8) , dimension(:), pointer :: indata_i8 , outdata_i8

    incol = column(intable, inColName, READ)
    outcol = addColumn(outTable, outColName, dataType(incol),units=unit, comment=comment)

    select case ( dataType(incol) )
    case (integer32)
       indata_i32  => int32Data( incol )
       outdata_i32 => int32Data( outcol )
       outdata_i32 = indata_i32( idx )
    case (integer16)
       indata_i16  => int16Data( incol )
       outdata_i16 => int16Data( outcol )
       outdata_i16 = indata_i16( idx )
    case (integer8)
       indata_i8   => int8Data( incol )
       outdata_i8  => int8Data( outcol )
       outdata_i8  = indata_i8( idx )
    case default
       write(errstr,*) 'Type', dataType(incol),'not implemented.'
       call error('selectedAddnCopyColumn:missingType')
    end select

    call release( incol )
    call release( outcol )

  end subroutine selectedAddnCopyColumn
end module rgsframesutils
